#include "StdAfx.h"
#include "DispatchProxy.h"
#include "log.h"

#include <strstream>
#include <ios>
#include <iomanip>


CDispatchProxy::CDispatchProxy():
    _patchedName("")
{
}

CDispatchProxy::~CDispatchProxy()
{
}


HRESULT CDispatchProxy::GetTypeInfoCount(UINT* pctinfo)
{
    COMSMART_PTR_CHECK(_pDispatch);
    return _pDispatch->GetTypeInfoCount(pctinfo);
}

HRESULT CDispatchProxy::GetTypeInfo(UINT itinfo, LCID lcid, ITypeInfo** pptinfo)
{
    COMSMART_PTR_CHECK(_pDispatch);
    return _pDispatch->GetTypeInfo(itinfo, lcid, pptinfo);
}

HRESULT CDispatchProxy::GetIDsOfNames(REFIID riid, LPOLESTR* rgszNames, UINT cNames,
                         LCID lcid, DISPID* rgdispid)
{
    COMSMART_PTR_CHECK(_pDispatch);
    return _pDispatch->GetIDsOfNames(riid, rgszNames, cNames, lcid, rgdispid);
}

HRESULT CDispatchProxy::Invoke(DISPID dispidMember, REFIID riid,
                  LCID lcid, WORD wFlags, DISPPARAMS* pdispparams, VARIANT* pvarResult,
                  EXCEPINFO* pexcepinfo, UINT* puArgErr)
{
    COMSMART_PTR_CHECK(_pDispatch);
    std::string methodName = getNameById(lcid, dispidMember);

    LOG4CXX_INFO(log::pMainLog, 
        "\n" << std::setw(17) << "method context - " << 
        " "  << "NAME: " << std::setw(15) << _patchedName << " " << 
        " "  << "ADDRESS: "<< std::setw(5) << std::hex << (void*) &*_pDispatch  << 
        "\n" << std::setw(17) << "method info - " << 
        " "  << "ID: "<< std::setw(5) << std::hex << dispidMember << 
        " "  << "NAME: " << methodName);


    HRESULT hres = S_OK;
    hres = _pDispatch->Invoke(
        dispidMember, riid, lcid, wFlags, pdispparams, pvarResult,pexcepinfo, puArgErr);
    if (FAILED(hres))
    {
        return hres;
    }
    for (UINT i = 0; i < pdispparams->cArgs; ++i)
    {
        if ((pdispparams->rgvarg[i].vt & (VT_BYREF | VT_DISPATCH)) == (VT_BYREF | VT_DISPATCH))
        {
            *(pdispparams->rgvarg[i].ppdispVal) = 
                dispatchProxy_t::createNewDispatchProxy(*(pdispparams->rgvarg[i].ppdispVal), methodName);
        }
    }

    if ((NULL != pvarResult) && 
        (VT_DISPATCH == pvarResult->vt) && 
        (NULL != pvarResult->pdispVal))
    {
        pvarResult->pdispVal = dispatchProxy_t::createNewDispatchProxy(pvarResult->pdispVal, methodName);
    }

    return hres;

}

void CDispatchProxy::setDisaptcher(IDispatch *pDispatch)
{
    _pDispatch = pDispatch;
}

IDispatch* CDispatchProxy::createNewDispatchProxy(IDispatch *pDispatch, std::string patchedName)
{

    CDispatchProxy *pProxy = new dispatchProxy_t();
    pProxy->AddRef();

    pProxy->setDisaptcher(pDispatch);

    LOG4CXX_INFO(log::pMainLog, 
        "dispatch was patched through " << 
        patchedName << 
        " call " << 
        std::setw(5) << std::hex << (pProxy));

    pProxy->_patchedName = patchedName;
    return pProxy;
}

std::string CDispatchProxy::getNameById(LCID lcid, DISPID id)
{
    CComPtr<ITypeInfo> pTypeInfo;
    HRESULT hres = S_OK;
    hres = _pDispatch->GetTypeInfo(0, lcid, &pTypeInfo);

    if (FAILED(hres))
    {
        return "";
    }

    BSTR bstrMethodName = NULL;
    UINT namesCount = 0;
    hres = pTypeInfo->GetNames(id, &bstrMethodName, 1, &namesCount);

    if (FAILED(hres))
    {
        return "";
    }

    char *ansiMethodName = _com_util::ConvertBSTRToString(bstrMethodName);
    std::string result = (NULL == ansiMethodName) ? "" : ansiMethodName;
    delete[] ansiMethodName;
    ::SysFreeString(bstrMethodName);
    return result;
}
